#ifndef THUNDER_ARM_H
#define THUNDER_ARM_H

#define ARM_TOLERANCE           6
#define ARM_MIN_SPEED_BASE		.0175
#define ARM_MIN_SPEED_MOD		.1
#define ARM_START_ANGLE			15 * ARM_ENCODING
#define ARM_MIN_STALL_SPEED		.12 //HORSE
#define ARM_MAX_SPEED			.75
#define ARM_RAMP_START			40 * ARM_ENCODING
#define ARM_DEPLOY_MIN_SPEED	.5
#define ARM_DEPLOY_MAX_SPEED	1
#define ARM_DEPLOY              5 * ARM_ENCODING
#define ARM_DEPLOY2				55 * ARM_ENCODING
#define ARM_DEPLOY3				ARM_DEPLOY2 + (5 * ARM_ENCODING)
#define ARM_CLOSE_TO_LOW		43 * ARM_ENCODING
#define ARM_TOO_LOW_FRONT		92
#define ARM_CLOSE_TO_HIGH		283 * ARM_ENCODING
#define ARM_TOO_LOW_BACK		305 * ARM_ENCODING


//NEVER CHANGE THIS TO FOUR. EVER. DON'T. LEAVE IT.
#define ARM_ENCODING			4

//Defines the zone behind the robot the arm can't be in
//The lower limit is the floor
#define ARM_MINI_ZONE			140 * ARM_ENCODING

//Defines the top of the zone in front of the robot the arm can't be in
//The lower limit is the floor
#define ARM_TUSK_ZONE			125 * ARM_ENCODING

//Defines the encoder number we must be below
	//to not hit the pole with the arm
#define ARM_POLE_ZONE			185 * ARM_ENCODING

//Defines the zone that the arm can't be in AFTER the minibot has deployed
	//Allows arm past the pole but not hitting the deployer
#define ARM_POST_MINI_ZONE		280 * ARM_ENCODING


#include "Resources.h"
#include "ThunderDeployer.h"
/*
 run the arm, manual control and presets for floor, low, mid, and high 
*/

class ThunderArm
{
public:
	ThunderArm(ThunderDeployer* deploypointer);
	~ThunderArm();
	
	typedef enum {PRE_NONE=0, 
		PRE_LOW, PRE_LOW_OFFSET, 
		PRE_MEDIUM, PRE_MEDIUM_OFFSET,
		PRE_HIGH, PRE_HIGH_OFFSET, 
		PRE_FEEDER, PRE_FLOOR, PRE_STRAIGHT_UP,
		PRE_AUTO_HIGH_OFFSET, PRE_AUTO_HIGH_OFFSET_LOWER,
		PRE_AUTO_MED_OFFSET, PRE_AUTO_MED_OFFSET_LOWER,
		PRE_AUTO_LOW_OFFSET, PRE_AUTO_LOW_OFFSET_LOWER,
		PRE_AUTO_HIGH, PRE_AUTO_HIGH_LOWER,
		PRE_AUTO_MED, PRE_AUTO_MED_LOWER,
		PRE_AUTO_LOW, PRE_AUTO_LOW_LOWER} PresetPositions;	
	
	
	//To barrel roll, press z or r twice
	
	//Does all the processing for the arm movement
	//MUST BE CALLED CONSTANTLY DURING AUTO AND TELE
	void ArmProcess();
	
	//Goes to a specific PRESET height
	//Selection determines which preset you go to
	//Inverted determines if the arm goes in front or behind the robot
		//(True = behind the robot, false = in front)
	//MUST BE CALLED WHEN WISHING TO CHANGE OR STOP THE PRESET
	void ArmGotoPreset(PresetPositions selection, bool inverted);
	
	//Sets the arm to try to stay where it is. 
	//Incompatible with anything but pre_none
	//Call this when you want to start
	void ArmSetAdhocPreset();
	
	//Stops the adhocing process
	//Should be called whenever the button is released
	//Call this when you want to stop	
	void ArmStopAdhoc();
	
	//Returns whether the arm is adhocing or not
	bool ArmGetAdhoc();
	
	//Disables the adhoc functionality of the arm
	void ArmDisableAdhoc(bool disable);
	
	//Gets whether adhocing is disabled
	bool ArmGetAdhocDisabled();
	
	//Tells the arm to go up or down manually (from the forward perspective)
	//Overridden by any preset operation
	//SHOULD NOT BE USED FOR AUTONOMOUS
	//MUST BE CALLED WHEN WISHING TO CHANGE OR STOP THE MANUAL SPEED
	void ArmGoUpDown(float speed);
	
	//Returns whether the arm is at the preset or not
	bool ArmCheckPresetDone();
	
	//Returns what preset the arm is currently set to
	PresetPositions ArmCheckPreset();
	
	//Returns whether the arm is in a position ready to deploy the minibot
	//Returns true for being ready and out of bad zones
	bool ArmCheckMinibotReady();
	
	//Returns whether the arm is in a position ready to deploy the tusks
	//Also used for minibot
	bool ArmCheckTusksReady();
	
	//CALL THIS AT THE END OF AUTONOMOUS
	void ArmResetAfterAuto();
	
	//Returns the encoder value that the arm is using
	int ArmDebugAngle();
	
	//Returns the actual encoder value for Left
	int ArmDebugEncoderLeft();
	
	//Returns the actual encoder value for Right
	int ArmDebugEncoderRight();
	
	//Returns the actual limit switch value
	bool ArmDebugFrontLimit();
	
	//Returns the actual limit switch value
	bool ArmDebugBackLimit();
	
	//Returns the actual limit switch value
	bool ArmDebugCalibLimit();
	
	//Sets whether we use encoder Left or encoder Right
	//false = encoder Left, true = encoder Right
	//(Just pass me the value of the switch)
	void ArmEncoderSwap(bool swap);
	
	//Sets whether both encoders are broken or not
	//True = they're both broken, false = at least 1 is fine
	void ArmEncoderBroken(bool broken);
	
	//Tells arm to fold out
	//Called ONCE and ONLY ONCE
	void ArmSetStartMatch();
	
	//Returns true if the arm is in the start match movement
	//Returns false once the movement is done
	bool ArmGetStartMatch();
	
	//Sets whether the calibration limit sensor is broken
	void ArmCalibLimitBroken(bool broken);
	
	//Sets whether the front limit switch is broken
	void ArmFrontLimitBroken(bool broken);
	
	//Sets whether the back limit switch is broken
	void ArmBackLimitBroken(bool broken);
	
	bool GetArmInverted();

	// return true if the arm cannot be used in autonomous
	bool IsArmBroken();
	
private:
	//Sees if "value" is within "goal" + or - "range"
	bool ArmWithinRange(float value, float range, float goal);
	
	//Checks for the arm being within range and is used to override motors in ArmProcess
	//Returns true if the arm is in an BAD position
	int ArmCheckOverrides();
	
	ThunderCANJaguar* armMotor;
	Encoder* armEncoderLeft;
	Encoder* armEncoderRight;
	AnalogChannel* armCalibLimit;
	DigitalInput* armFrontLimit;
	DigitalInput* armBackLimit;
	ThunderDeployer* deployer;
	float armSpeed; 
	ThunderArm::PresetPositions preset;
	float preset_angle;
	int actual_angle;
	int adhoc_angle;
	bool adhocing;
	bool adhoc_disabled;
	bool invert;
	bool preset_done;
	float presetData [22];
	float presetDataInverted [10];
	
	bool encoder_swapped;
	bool encoder_broken;
	bool start_match;
	
	bool calib_limit_broken;
	bool front_limit_broken;
	bool back_limit_broken;
	
	int angle_error_Left;
	int angle_error_Right;
	
	float ARM_MIN_SPEED;
};

#endif
