#ifndef THUNDER_DRIVE_H
#define THUNDER_DRIVE_H

#include "Resources.h"
#include "Tracker.h"

// size of line tracking state history
#define HIST_SIZE 1024
#define NUM_CIMS 4
/*
drive, low level autonomous, follow line
*/

class ThunderDrive
{
private:
	
//object pointers
	 //speed controllers
	 ThunderCANJaguar * motor[NUM_CIMS];
	 //encoders, used for distance as well as sanity checks when following the line
	 Counter * encoder[2];
	 //used for dead reckonning angle turns, as well as sanity checks when following the line
	 Gyro * thunder_gyro;
	 // Line sensor array for autonomous line tracking
	 Tracker * tracker;
	 //used for encoder reseting and speed calculations
	 Timer * speed_timer;
// \object pointers
	 

	 
//autonomous members
	 //type definition for the last called automatic task
	 typedef enum auto_task {drive_dead_reckonning, drive_using_line,
		                     turn_dead_reckonning, turn_using_line, none};
	 auto_task current_auto_task;
	 
	 //used for keeping robot straight in dead reckonning
	 float auto_start_angle;
	 //current speed after autonomous ramping is applied
	 float current_auto_speed;
	 float max_auto_speed;
	 float last_distance;
	 int ramp_state; //0:ramp up (until either halfway or max speed, whichever first)
	 				 //1: staying at max speed (until min stop dist is met)
	 				 //2: ramping down (stop when dist done)
	 static const float ramp_rate = 0.001;
	 //reference values for automatic tasks
	 double auto_start;
	 double auto_progress; 
	 double auto_goal;
// \autonomous members
	 
//line tracking members
	 //set when final design for sensors is made
	 int number_of_light_sensors;
 	 // history of what has happened, line tracking-wise; this is a ring
	 // buffer -- that is, when it's full, recording begins again
	 // from the beginning, overwriting previous values 
	 int history[HIST_SIZE];
	 int histpoint; // index in history
	 // do we need to find the line first, or are already on it?
	 bool fromScratch;
	 bool onLine; // self-explanatory
	 bool endOfLine; // have we reached it yet?
	 // max speed when we don't have enough distance to do the full
	 // speed curve usually done by Get_Max_Line_Speed
	 float speed_curve_k;
	 // factor to slow down acceleration
	 float ramp_factor;
	 // tracker state (semi-left or semi-right) at which to stop
	 Tracker::TrackState turn_goal;
// \line tracking members	 

//misc private members
	 //last set speeds
	 float l_speed;
	 float r_speed;
	 
	 float current_speed;
// \misc private member
	 
	 
//private utility methods	 
	 //number of times to rotate the wheel to move given inches
	 float Inches_To_Revolutions(float inches);
	 //number of inches moved given the number of ticks 
	 double Ticks_To_Inches(int ticks);
	 //monitor the line sensors, return motor output values
	 // to maintain straight path
	 void Stay_On_Line(float& leftSpeed, float& rightSpeed);
	 //helper function for Stay_On_Line; taking max_auto_speed
	 // and distance traveled (auto_progress), calculates
	 // optimum motor speed assuming robot is going straight
	 // (angle correction in Stay_On_Line)
	 float Get_Max_Line_Speed();

	 //apply a ramping value to motor speeds based on goal, progress, maximum, and ramping values
	 void Apply_Ramping(float * current_speed, float progress, float goal,
			 float up_rate, float down_rate, float max, float stop_up, 
			 float start_down, float * L, float * R);
	 
	 //apply gyro correction to the left and right motors
	 void Apply_Gyro_Correction(float target, float current, float factor, float * L, float * R);
	 
	 //routinely output various debug values for dead reckonning. disable by changing DEAD_RECKONING_DEBUG to 0
	 void Dead_Reckonning_Debug();
// \private utility methods
// \private:
public:
	
//class maintenance
	//constructor, set all variables to defaut, create gyro, motors, and encoders
	ThunderDrive();
	//destructor, delete pointers
	~ThunderDrive();
	//resets the gyro to 0 degrees,
	// do not call during autonomous motion
	void Drive_Clear_Gyro();
	
	void Set_Coast(bool coast);
// \class maintenance
	
//Teleoperated Methods
	//basic teleoperated function, set left and right motors speeds
		//-1 full reverse, 1 full forward, 0 stopped.
	void Drive_Left_Right_Motors(float left_speed, float right_speed);
// \Teleoperated Methods
	
//Debug Functions
	
	//Get the last passed left and right motor values, either from an outside call or an automatic task
	float Get_Last_Left_Speed();
	float Get_Last_Right_Speed();
	//use the encoders to determine current robot velocity. 
		//determined by an encoder read every call to the main teleop drive function.
		//((average encoder ticks / ticks per revolution) * inches per revolution) / time between loops (0.005 seconds)
	float Get_Current_Robot_Speed();
	
	//the last called automatic task
	auto_task Get_Current_Auto_Task();
	
	//the last called automatic task progress
		//Drive_Distance/Keep_On_Line: inches driven from start
		//Turn_Angle/Turn_Until_Line: change in degrees turned since start
	float Get_Auto_Progress();
	//the last called automatic task start
		//Drive_Distance/Keep_On_Line: encoder distance
		//Turn_Angle/Turn_Until_Line: gyro angle at call
	float Get_Auto_Start();
	
	//the last called automatic task goal
		//Drive_Distance/Keep_On_Line: inches to drive
	float Get_Auto_Goal();
	
	//returns whether or not the sensors indicate the robot is positioned on a line. 
	bool  Get_Is_On_Line();
	
	 //check the given light sensor.
		//0: left sensor. 1: center sensor. 2: right sensor
	 	//true: on line
	 	//false: not on line
	bool Sensor_Active(int sensor);
	
	//checks an array (size of 3) of boolean values against the values of the 3 light sensors.  see Sensor_Active for conventions. 
	bool Light_Sensors_Match(bool * sensor);

	//returns the number of encoder ticks since last update, true 
	//for left encoder, false for right encoder
	//returns both positive and negative (forwards and backwards)
	int Get_Encoder_Ticks(bool leftTrightF);
	
	//returns angle the gyro is currently at
	//Range: Unbounded in degrees
	float Get_Gyro_Angle();
// \Debug Functions
	
//Low Level Autonomous:
	//General Structure:
	//1. Give Task (Drive_Forward_Distance, Drive_Keep_On_Line, Drive_Turn_Angle, Drive_Turn_Until_Line)
	//2. Process (Process_Auto() )
	//3. Check Done (Check_Auto_Done() )
	//4. if not done, loop to 2, if done, continue
	//5. loop to 1 until all tasks are completed
	
	
	//sets the fastest motor speed that can be output during an automatic task. ramping up and down
		//will depend on the value.
		//valid values are 0 to 1, range resulting will be [-max_speed, max_speed] 
	void Drive_Set_Max_Auto_Speed(float max_speed);
	

	
//Automatic Tasks:	(Note: can be used outside of autonomous, do not call Drive_Left_Right_Motors
							//unless the action is completed or canceled using Stop_Auto_Task)
	
	//starts a task to: drive the provided number of inches, negative results in backwards driving. 
		//attempts to stay straight.
	void Drive_Forward_Distance(float inches);
	
	//starts a task to:turn the robot the given number of degrees relative to the current angle,
		//positive is to CW, negative is CCW, uses degrees
	void Drive_Turn_Angle(float degrees);
	
	//starts a task to:maintain the robot on the line while driving forward, used for 2" lines.
	//the robot continues until the robot is no longer over a single line or the specified distance
		//is driven.
		//if init is true, robot will make sure it's on the line before proceeding
		//if init is false, it will assume it's already on the line
	void Drive_Keep_On_Line(float inches = 9000.01, bool init = true);
	
	//starts a task to:turn the robot until it centers over the specified line or the specified 
		//number of degrees has been turned. see Drive_Turn_Angle function for conventions.
	void Drive_Turn_Until_Line(float degrees);
	
	//stop the current automatic task. Does not matter if there is a task in progress.
	void Drive_Stop_Auto_Task();
	
// \Automatic Tasks

//Automatic Loop Methods
	//call every loop, when running an automatic task, to run all necessary internal functions
	void Drive_Process_Auto();
	
	//returns true when the last called autonomous task is complete, Conditions:
		//Drive_Distance: current encoder ticks indicate the distance has been driven
		//Turn_Angle: gyro current equals gyro start plus given angle
		//Keep_On_Line: sensors indicate there is no more line in front of the robot, or the distance
			//specified
		//Turn_Until_Line: sensors indicate the robot has turned and is now aligned with a line
			//or the maximum angle has been driven.
	bool Drive_Current_Auto_Task_Done(); 
// \Automatic Loop Methods

// \Low Level Autonomous
	
// \public:
	
}; // \class ThunderDrive

#endif // \#ifndef THUNDER_DRIVE_H

