#include "WPILib.h"
#include "Tracker.h"

Tracker::Tracker(int leftChannel, int centerChannel, int rightChannel) :
	sLeft(leftChannel), sCenter(centerChannel), sRight(rightChannel)
{
	totalStates = 0;
}

// The most important function here:
// WHY do I need Tracker:: in front of TrackState here??
Tracker::TrackState Tracker::GetState() {
	if (sCenter.GetVoltage() > V_HIGH) {
		if (sLeft.GetVoltage() < V_LOW) {
			// 0 - 1 - 0 : on the line
			if(sRight.GetVoltage() < V_LOW)
				return ON_LINE;
			// 0 - 1 - 1 : slightly too far left
			else
				return SEMI_LEFT;
		}
		else { // assume left is low
			// 1 - 1 - 1 : at stop bar (or with low probability at wye)
			if (sRight.GetVoltage() > V_HIGH)
				return STOP;
			// 1 - 1 - 0 : slightly too far right
			else 
				return SEMI_RIGHT;
		}
	} else { // assume center is low
		if (sLeft.GetVoltage() > V_HIGH) {
			// 1 - 0 - 0 : too far right
			if(sRight.GetVoltage() < V_LOW)
				return TOO_RIGHT;
			// 1 - 0 - 1 : must be at wye
			else
				return WYE;
		} else { // assume it's low as well
			// 0 - 0 - 1 : too far left
			if(sRight.GetVoltage() > V_HIGH)
				return TOO_LEFT;
			// 0 - 0 - 0 : completely off line--big trouble!
			else
				return WAY_OFF;
		}
	}
}

// Buffered version of GetState()
Tracker::TrackState Tracker::GetBufferedState() {
	float left = 0, center = 0, right = 0;
	// move each buffer down one
	for(int i = 0; i<BUFFER_SIZE-1; i++) {
		pastLeft[i] = pastLeft[i+1];
		left += pastLeft[i];
		pastCenter[i] = pastCenter[i+1];
		center += pastCenter[i];
		pastRight[i] = pastRight[i+1];
		right += pastRight[i];
	}
	// get newest values and store as on/off
	pastLeft[BUFFER_SIZE-1] = (sLeft.GetVoltage() > V_HIGH) ? 1 : 0;
	pastCenter[BUFFER_SIZE-1] = (sCenter.GetVoltage() > V_HIGH) ? 1 : 0;
	pastRight[BUFFER_SIZE-1] = (sRight.GetVoltage() > V_HIGH) ? 1 : 0;
	// have we filled the buffer yet? if so, we can proceed
	if(totalStates >= BUFFER_SIZE) { 
		left += pastLeft[BUFFER_SIZE-1];
		center += pastCenter[BUFFER_SIZE-1];
		right += pastRight[BUFFER_SIZE-1];
		// take average of past values
	    left /= BUFFER_SIZE;
	    center /= BUFFER_SIZE;
	    right /= BUFFER_SIZE;
	} else {
		// don't have enough values, just take the present values
		left = pastLeft[BUFFER_SIZE-1];
		center = pastCenter[BUFFER_SIZE-1];
		right = pastRight[BUFFER_SIZE-1];
	}
	// center more on than not
	if (center > 0.5) {
		if (left < 0.5) {
			// 0 - 1 - 0 : on the line
			if(right < 0.5)
				return ON_LINE;
			// 0 - 1 - 1 : slightly too far left
			else
				return SEMI_LEFT;
		}
		else { // assume left is low
			// 1 - 1 - 1 : at stop bar (or with low probability at wye)
			if (right > 0.5)
				return STOP;
			// 1 - 1 - 0 : slightly too far right
			else 
				return SEMI_RIGHT;
		}
	} else { // assume center is low
		if (left > 0.5) {
			// 1 - 0 - 0 : too far right
			if(right < 0.5)
				return TOO_RIGHT;
			// 1 - 0 - 1 : must be at wye
			else
				return WYE;
		} else { // assume it's low as well
			// 0 - 0 - 1 : too far left
			if(right > 0.5)
				return TOO_LEFT;
			// 0 - 0 - 0 : completely off line--big trouble!
			else
				return WAY_OFF;
		}
	}
}


// debug output functions
bool Tracker::GetLeft() {
	return (sLeft.GetVoltage() > V_HIGH);
}
bool Tracker::GetCenter() {
	return (sCenter.GetVoltage() > V_HIGH);
}
bool Tracker::GetRight() {
	return (sRight.GetVoltage() > V_HIGH);
}
void Tracker::PrintState(TrackState state) {
	printf("Tracker is ");
	if(state==ON_LINE) printf("ON_LINE\n");
	if(state==SEMI_LEFT) printf("SEMI_LEFT\n");
	if(state==SEMI_RIGHT) printf("SEMI_RIGHT\n");
	if(state==WAY_OFF) printf("WAY_OFF\n");
	if(state==TOO_LEFT) printf("TOO_LEFT\n");
	if(state==TOO_RIGHT) printf("TOO_RIGHT\n");
	if(state==STOP) printf("STOP\n");
	if(state==WYE) printf("WYE\n");
}
